import uuid
from datetime import datetime, timedelta, timezone
from typing import AsyncGenerator

import pytest
from fastapi_users.authentication.strategy.db.models import BaseAccessToken
from pydantic import UUID4
from sqlalchemy import exc
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, create_async_engine
from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base
from sqlalchemy.orm import sessionmaker

from fastapi_users_db_sqlalchemy import SQLAlchemyBaseUserTable
from fastapi_users_db_sqlalchemy.access_token import (
    SQLAlchemyAccessTokenDatabase,
    SQLAlchemyBaseAccessTokenTable,
)


class AccessToken(BaseAccessToken):
    pass


def create_async_session_maker(engine: AsyncEngine):
    return sessionmaker(engine, class_=AsyncSession, expire_on_commit=False)


@pytest.fixture
def user_id() -> UUID4:
    return uuid.uuid4()


@pytest.fixture
async def sqlalchemy_access_token_db(
    user_id: UUID4,
) -> AsyncGenerator[SQLAlchemyAccessTokenDatabase, None]:
    Base: DeclarativeMeta = declarative_base()

    class AccessTokenTable(SQLAlchemyBaseAccessTokenTable, Base):
        pass

    class UserTable(SQLAlchemyBaseUserTable, Base):
        pass

    DATABASE_URL = "sqlite+aiosqlite:///./test-sqlalchemy-access-token.db"
    engine = create_async_engine(
        DATABASE_URL, connect_args={"check_same_thread": False}
    )
    sessionmaker = create_async_session_maker(engine)

    async with engine.begin() as connection:
        await connection.run_sync(Base.metadata.create_all)

        async with sessionmaker() as session:
            user = UserTable(
                id=user_id, email="lancelot@camelot.bt", hashed_password="guinevere"
            )
            session.add(user)
            await session.commit()

            yield SQLAlchemyAccessTokenDatabase(AccessToken, session, AccessTokenTable)

        await connection.run_sync(Base.metadata.drop_all)


@pytest.mark.asyncio
@pytest.mark.db
async def test_queries(
    sqlalchemy_access_token_db: SQLAlchemyAccessTokenDatabase[AccessToken],
    user_id: UUID4,
):
    access_token = AccessToken(token="TOKEN", user_id=user_id)

    # Create
    access_token_db = await sqlalchemy_access_token_db.create(access_token)
    assert access_token_db.token == "TOKEN"
    assert access_token_db.user_id == user_id

    # Update
    access_token_db.created_at = datetime.now(timezone.utc)
    await sqlalchemy_access_token_db.update(access_token_db)

    # Get by token
    access_token_by_token = await sqlalchemy_access_token_db.get_by_token(
        access_token_db.token
    )
    assert access_token_by_token is not None

    # Get by token expired
    access_token_by_token = await sqlalchemy_access_token_db.get_by_token(
        access_token_db.token, max_age=datetime.now(timezone.utc) + timedelta(hours=1)
    )
    assert access_token_by_token is None

    # Get by token not expired
    access_token_by_token = await sqlalchemy_access_token_db.get_by_token(
        access_token_db.token, max_age=datetime.now(timezone.utc) - timedelta(hours=1)
    )
    assert access_token_by_token is not None

    # Get by token unknown
    access_token_by_token = await sqlalchemy_access_token_db.get_by_token(
        "NOT_EXISTING_TOKEN"
    )
    assert access_token_by_token is None

    # Delete token
    await sqlalchemy_access_token_db.delete(access_token_db)
    deleted_access_token = await sqlalchemy_access_token_db.get_by_token(
        access_token_db.token
    )
    assert deleted_access_token is None


@pytest.mark.asyncio
@pytest.mark.db
async def test_insert_existing_token(
    sqlalchemy_access_token_db: SQLAlchemyAccessTokenDatabase[AccessToken],
    user_id: UUID4,
):
    access_token = AccessToken(token="TOKEN", user_id=user_id)
    await sqlalchemy_access_token_db.create(access_token)

    with pytest.raises(exc.IntegrityError):
        await sqlalchemy_access_token_db.create(
            AccessToken(token="TOKEN", user_id=user_id)
        )



import asyncio
from typing import Optional

import pytest
from fastapi_users import models


class User(models.BaseUser):
    first_name: Optional[str]


class UserCreate(models.BaseUserCreate):
    first_name: Optional[str]


class UserUpdate(models.BaseUserUpdate):
    pass


class UserDB(User, models.BaseUserDB):
    pass


class UserOAuth(User, models.BaseOAuthAccountMixin):
    pass


class UserDBOAuth(UserOAuth, UserDB):
    pass


@pytest.fixture(scope="session")
def event_loop():
    """Force the pytest-asyncio loop to be the main one."""
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture
def oauth_account1() -> models.BaseOAuthAccount:
    return models.BaseOAuthAccount(
        oauth_name="service1",
        access_token="TOKEN",
        expires_at=1579000751,
        account_id="user_oauth1",
        account_email="king.arthur@camelot.bt",
    )


@pytest.fixture
def oauth_account2() -> models.BaseOAuthAccount:
    return models.BaseOAuthAccount(
        oauth_name="service2",
        access_token="TOKEN",
        expires_at=1579000751,
        account_id="user_oauth2",
        account_email="king.arthur@camelot.bt",
    )



import uuid

from pydantic import UUID4
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.types import CHAR, TypeDecorator


class GUID(TypeDecorator):  # pragma: no cover
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(36), storing as regular strings.
    """

    class UUIDChar(CHAR):
        python_type = UUID4

    impl = UUIDChar
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(36))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return str(uuid.UUID(value))
            else:
                return str(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value



"""FastAPI Users database adapter for SQLAlchemy."""
from typing import Optional, Type

from fastapi_users.db.base import BaseUserDatabase
from fastapi_users.models import UD
from pydantic import UUID4
from sqlalchemy import (
    Boolean,
    Column,
    ForeignKey,
    Integer,
    String,
    delete,
    func,
    select,
    update,
)
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import joinedload
from sqlalchemy.sql import Select

from fastapi_users_db_sqlalchemy.guid import GUID

__version__ = "3.0.1"


class SQLAlchemyBaseUserTable:
    """Base SQLAlchemy users table definition."""

    __tablename__ = "user"

    id = Column(GUID, primary_key=True)
    email = Column(String(length=320), unique=True, index=True, nullable=False)
    hashed_password = Column(String(length=1024), nullable=False)
    is_active = Column(Boolean, default=True, nullable=False)
    is_superuser = Column(Boolean, default=False, nullable=False)
    is_verified = Column(Boolean, default=False, nullable=False)


class SQLAlchemyBaseOAuthAccountTable:
    """Base SQLAlchemy OAuth account table definition."""

    __tablename__ = "oauth_account"

    id = Column(GUID, primary_key=True)
    oauth_name = Column(String(length=100), index=True, nullable=False)
    access_token = Column(String(length=1024), nullable=False)
    expires_at = Column(Integer, nullable=True)
    refresh_token = Column(String(length=1024), nullable=True)
    account_id = Column(String(length=320), index=True, nullable=False)
    account_email = Column(String(length=320), nullable=False)

    @declared_attr
    def user_id(cls):
        return Column(GUID, ForeignKey("user.id", ondelete="cascade"), nullable=False)


class SQLAlchemyUserDatabase(BaseUserDatabase[UD]):
    """
    Database adapter for SQLAlchemy.

    :param user_db_model: Pydantic model of a DB representation of a user.
    :param session: SQLAlchemy session instance.
    :param user_table: SQLAlchemy user model.
    :param oauth_account_table: Optional SQLAlchemy OAuth accounts model.
    """

    session: AsyncSession
    user_table: Type[SQLAlchemyBaseUserTable]
    oauth_account_table: Optional[Type[SQLAlchemyBaseOAuthAccountTable]]

    def __init__(
        self,
        user_db_model: Type[UD],
        session: AsyncSession,
        user_table: Type[SQLAlchemyBaseUserTable],
        oauth_account_table: Optional[Type[SQLAlchemyBaseOAuthAccountTable]] = None,
    ):
        super().__init__(user_db_model)
        self.session = session
        self.user_table = user_table
        self.oauth_account_table = oauth_account_table

    async def get(self, id: UUID4) -> Optional[UD]:
        statement = select(self.user_table).where(self.user_table.id == id)
        return await self._get_user(statement)

    async def get_by_email(self, email: str) -> Optional[UD]:
        statement = select(self.user_table).where(
            func.lower(self.user_table.email) == func.lower(email)
        )
        return await self._get_user(statement)

    async def get_by_oauth_account(self, oauth: str, account_id: str) -> Optional[UD]:
        if self.oauth_account_table is not None:
            statement = (
                select(self.user_table)
                .join(self.oauth_account_table)
                .where(self.oauth_account_table.oauth_name == oauth)
                .where(self.oauth_account_table.account_id == account_id)
            )
            return await self._get_user(statement)

    async def create(self, user: UD) -> UD:
        user_table = self.user_table(**user.dict(exclude={"oauth_accounts"}))
        self.session.add(user_table)

        if self.oauth_account_table is not None:
            for oauth_account in user.oauth_accounts:
                oauth_account_table = self.oauth_account_table(
                    **oauth_account.dict(), user_id=user.id
                )
                self.session.add(oauth_account_table)

        await self.session.commit()
        return await self.get(user.id)

    async def update(self, user: UD) -> UD:
        user_table = await self.session.get(self.user_table, user.id)
        for key, value in user.dict(exclude={"oauth_accounts"}).items():
            setattr(user_table, key, value)
        self.session.add(user_table)

        if self.oauth_account_table is not None:
            for oauth_account in user.oauth_accounts:
                statement = update(
                    self.oauth_account_table,
                    whereclause=self.oauth_account_table.id == oauth_account.id,
                    values={**oauth_account.dict(), "user_id": user.id},
                )
                await self.session.execute(statement)

        await self.session.commit()

        return await self.get(user.id)

    async def delete(self, user: UD) -> None:
        statement = delete(self.user_table, self.user_table.id == user.id)
        await self.session.execute(statement)
        await self.session.commit()

    async def _get_user(self, statement: Select) -> Optional[UD]:
        if self.oauth_account_table is not None:
            statement = statement.options(joinedload("oauth_accounts"))

        results = await self.session.execute(statement)
        user = results.first()
        if user is None:
            return None

        return self.user_db_model.from_orm(user[0])



from typing import AsyncGenerator

import pytest
from sqlalchemy import Column, String, exc
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, create_async_engine
from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base
from sqlalchemy.orm import relationship, sessionmaker

from fastapi_users_db_sqlalchemy import (
    SQLAlchemyBaseOAuthAccountTable,
    SQLAlchemyBaseUserTable,
    SQLAlchemyUserDatabase,
)
from tests.conftest import UserDB, UserDBOAuth


def create_async_session_maker(engine: AsyncEngine):
    return sessionmaker(engine, class_=AsyncSession, expire_on_commit=False)


@pytest.fixture
async def sqlalchemy_user_db() -> AsyncGenerator[SQLAlchemyUserDatabase, None]:
    Base: DeclarativeMeta = declarative_base()

    class User(SQLAlchemyBaseUserTable, Base):
        first_name = Column(String, nullable=True)

    DATABASE_URL = "sqlite+aiosqlite:///./test-sqlalchemy-user.db"

    engine = create_async_engine(
        DATABASE_URL, connect_args={"check_same_thread": False}
    )
    sessionmaker = create_async_session_maker(engine)

    async with engine.begin() as connection:
        await connection.run_sync(Base.metadata.create_all)

        async with sessionmaker() as session:
            yield SQLAlchemyUserDatabase(UserDB, session, User)

        await connection.run_sync(Base.metadata.drop_all)


@pytest.fixture
async def sqlalchemy_user_db_oauth() -> AsyncGenerator[SQLAlchemyUserDatabase, None]:
    Base: DeclarativeMeta = declarative_base()

    class User(SQLAlchemyBaseUserTable, Base):
        first_name = Column(String, nullable=True)
        oauth_accounts = relationship("OAuthAccount")

    class OAuthAccount(SQLAlchemyBaseOAuthAccountTable, Base):
        pass

    DATABASE_URL = "sqlite+aiosqlite:///./test-sqlalchemy-user-oauth.db"

    engine = create_async_engine(
        DATABASE_URL, connect_args={"check_same_thread": False}
    )
    sessionmaker = create_async_session_maker(engine)

    async with engine.begin() as connection:
        await connection.run_sync(Base.metadata.create_all)

        async with sessionmaker() as session:
            yield SQLAlchemyUserDatabase(UserDBOAuth, session, User, OAuthAccount)

        await connection.run_sync(Base.metadata.drop_all)


@pytest.mark.asyncio
@pytest.mark.db
async def test_queries(sqlalchemy_user_db: SQLAlchemyUserDatabase[UserDB]):
    user = UserDB(
        email="lancelot@camelot.bt",
        hashed_password="guinevere",
    )

    # Create
    user_db = await sqlalchemy_user_db.create(user)
    assert user_db.id is not None
    assert user_db.is_active is True
    assert user_db.is_superuser is False
    assert user_db.email == user.email

    # Update
    user_db.is_superuser = True
    await sqlalchemy_user_db.update(user_db)

    # Get by id
    id_user = await sqlalchemy_user_db.get(user.id)
    assert id_user is not None
    assert id_user.id == user_db.id
    assert id_user.is_superuser is True

    # Get by email
    email_user = await sqlalchemy_user_db.get_by_email(str(user.email))
    assert email_user is not None
    assert email_user.id == user_db.id

    # Get by uppercased email
    email_user = await sqlalchemy_user_db.get_by_email("Lancelot@camelot.bt")
    assert email_user is not None
    assert email_user.id == user_db.id

    # Unknown user
    unknown_user = await sqlalchemy_user_db.get_by_email("galahad@camelot.bt")
    assert unknown_user is None

    # Delete user
    await sqlalchemy_user_db.delete(user)
    deleted_user = await sqlalchemy_user_db.get(user.id)
    assert deleted_user is None


@pytest.mark.asyncio
@pytest.mark.db
async def test_insert_existing_email(
    sqlalchemy_user_db: SQLAlchemyUserDatabase[UserDB],
):
    user = UserDB(
        email="lancelot@camelot.bt",
        hashed_password="guinevere",
    )
    await sqlalchemy_user_db.create(user)

    with pytest.raises(exc.IntegrityError):
        await sqlalchemy_user_db.create(
            UserDB(email=user.email, hashed_password="guinevere")
        )


@pytest.mark.asyncio
@pytest.mark.db
async def test_queries_custom_fields(
    sqlalchemy_user_db: SQLAlchemyUserDatabase[UserDB],
):
    """It should output custom fields in query result."""
    user = UserDB(
        email="lancelot@camelot.bt",
        hashed_password="guinevere",
        first_name="Lancelot",
    )
    await sqlalchemy_user_db.create(user)

    id_user = await sqlalchemy_user_db.get(user.id)
    assert id_user is not None
    assert id_user.id == user.id
    assert id_user.first_name == user.first_name


@pytest.mark.asyncio
@pytest.mark.db
async def test_queries_oauth(
    sqlalchemy_user_db_oauth: SQLAlchemyUserDatabase[UserDBOAuth],
    oauth_account1,
    oauth_account2,
):
    user = UserDBOAuth(
        email="lancelot@camelot.bt",
        hashed_password="guinevere",
        oauth_accounts=[oauth_account1, oauth_account2],
    )

    # Create
    user_db = await sqlalchemy_user_db_oauth.create(user)
    assert user_db.id is not None
    assert hasattr(user_db, "oauth_accounts")
    assert len(user_db.oauth_accounts) == 2

    # Update
    user_db.oauth_accounts[0].access_token = "NEW_TOKEN"
    await sqlalchemy_user_db_oauth.update(user_db)

    # Get by id
    id_user = await sqlalchemy_user_db_oauth.get(user.id)
    assert id_user is not None
    assert id_user.id == user_db.id
    assert id_user.oauth_accounts[0].access_token == "NEW_TOKEN"

    # Get by email
    email_user = await sqlalchemy_user_db_oauth.get_by_email(str(user.email))
    assert email_user is not None
    assert email_user.id == user_db.id
    assert len(email_user.oauth_accounts) == 2

    # Get by OAuth account
    oauth_user = await sqlalchemy_user_db_oauth.get_by_oauth_account(
        oauth_account1.oauth_name, oauth_account1.account_id
    )
    assert oauth_user is not None
    assert oauth_user.id == user.id

    # Unknown OAuth account
    unknown_oauth_user = await sqlalchemy_user_db_oauth.get_by_oauth_account(
        "foo", "bar"
    )
    assert unknown_oauth_user is None



from datetime import datetime
from typing import Generic, Optional, Type

from fastapi_users.authentication.strategy.db import A, AccessTokenDatabase
from sqlalchemy import Column, DateTime, ForeignKey, String, delete, select, update
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.declarative import declared_attr

from fastapi_users_db_sqlalchemy.guid import GUID


class SQLAlchemyBaseAccessTokenTable:
    """Base SQLAlchemy access token table definition."""

    __tablename__ = "accesstoken"

    token = Column(String(length=43), primary_key=True)
    created_at = Column(DateTime(timezone=True), index=True, nullable=False)

    @declared_attr
    def user_id(cls):
        return Column(GUID, ForeignKey("user.id", ondelete="cascade"), nullable=False)


class SQLAlchemyAccessTokenDatabase(AccessTokenDatabase, Generic[A]):
    """
    Access token database adapter for SQLAlchemy.

    :param access_token_model: Pydantic model of a DB representation of an access token.
    :param session: SQLAlchemy session instance.
    :param access_token_table: SQLAlchemy access token model.
    """

    def __init__(
        self,
        access_token_model: Type[A],
        session: AsyncSession,
        access_token_table: Type[SQLAlchemyBaseAccessTokenTable],
    ):
        self.access_token_model = access_token_model
        self.session = session
        self.access_token_table = access_token_table

    async def get_by_token(
        self, token: str, max_age: Optional[datetime] = None
    ) -> Optional[A]:
        statement = select(self.access_token_table).where(
            self.access_token_table.token == token
        )
        if max_age is not None:
            statement = statement.where(self.access_token_table.created_at >= max_age)

        results = await self.session.execute(statement)
        access_token = results.first()
        if access_token is None:
            return None
        return self.access_token_model.from_orm(access_token[0])

    async def create(self, access_token: A) -> A:
        access_token_db = self.access_token_table(**access_token.dict())
        self.session.add(access_token_db)
        await self.session.commit()
        return access_token

    async def update(self, access_token: A) -> A:
        statement = (
            update(self.access_token_table)
            .where(self.access_token_table.token == access_token.token)
            .values(access_token.dict())
        )
        await self.session.execute(statement)
        await self.session.commit()
        return access_token

    async def delete(self, access_token: A) -> None:
        statement = delete(
            self.access_token_table, self.access_token_table.token == access_token.token
        )
        await self.session.execute(statement)
        await self.session.commit()
